/* gmpromisify v0.1.8 - https://github.com/syusui-s/gmpromisify/
 *
 *   Copyright 2022-2025  Shusui Moyatani
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */
var gmPromisify = (function (exports) {
    'use strict';

    const buildDetails$1 = (args) => {
        if (typeof args[0] === 'string' || args[0] instanceof URL) {
            const url = args[0].toString();
            const name = args[1];
            if (name == null) {
                throw new Error('name is empty.');
            }
            return { url, name };
        }
        return args[0];
    };
    /**
     * GM_download wrapper.
     */
    const gmDownload = (...args) => {
        if (typeof globalThis.GM_download === 'undefined') {
            throw new Error('GM_download is not listed in @grant or not supported on this environment.');
        }
        const { signal, ...details } = buildDetails$1(args);
        if (signal?.aborted) {
            throw new DOMException('The request was aborted.', 'AbortError');
        }
        return new Promise((resolve, reject) => {
            const control = GM_download({
                ...details,
                onload() {
                    resolve();
                },
                onerror(err) {
                    const errDetails = err.details != null ? `: ${err.details}` : '';
                    reject(new TypeError(`DownloadError: ${err.error}${errDetails}`));
                },
                ontimeout() {
                    reject(new TypeError('TimeoutError'));
                },
            });
            if (signal != null) {
                signal.addEventListener('abort', () => control.abort());
            }
        });
    };

    /**
     * Encode binary data as isomorphic-encoded string.
     *
     * Binary data cannot be encoded as JavaScript string directly
     * because some binary data may contain some binary sequences
     * which are treated specially in the text encoding.
     * GM.xmlhttpRequest requires that each single byte (uint8) is encoded as CharCode.
     * By doing this, it is possible for GM.xmlHttpRequest to obtain the original binary data.
     *
     * https://developer.mozilla.org/en-US/docs/Web/API/btoa
     * https://infra.spec.whatwg.org/#isomorphic-encode
     * https://infra.spec.whatwg.org/#isomorphic-decode
     *
     * GreaseMonkey
     *   https://github.com/greasemonkey/greasemonkey/blob/efd22a93121225ada47f2fe9f021af0ab6100c21/src/bg/on-user-script-xhr.js#L140-L146
     */
    const arrayBufferToString = (b) => {
        const view = new Uint8Array(b);
        const batch = 1024;
        const result = [];
        for (let i = 0; i < view.length; i += batch) {
            const str = String.fromCharCode(...view.subarray(i, i + batch));
            result.push(str);
        }
        return result.join('');
    };

    /**
     * Parse response headers string and covert it to Headers object.
     *
     * https://developer.mozilla.org/ja/docs/Web/API/XMLHttpRequest/getAllResponseHeaders
     */
    const parseXHRHeaders = (gmHeaders) => {
        const headers = new Headers();
        if (gmHeaders == null) {
            return headers;
        }
        // These `name` and `value` contains the last value
        let name = '';
        let value = '';
        // According to 2.2 [RFC9112], a recipient may recognize LF(\n) as a line terminator and ignore CR(\r).
        // However, this implementation strictly requires a sequence CRLF('\r\n') for Firefox-styled multi-lined Set-Cookie.
        // Firefox XHR returns multiple Set-Cookie as LF(\n) separated lines.
        gmHeaders.split(/(\r\n)+/).forEach((line) => {
            const index = line.indexOf(':');
            // The line with no colon (:) is ignored.
            if (index < 0) {
                // According to 3.2.2 [RFC9112], obs-fold should be replaced with SP
                if (/^[ \t]/.test(line) && name.length > 0 && value.length > 0) {
                    const appendValue = line.trim();
                    headers.set(name, `${value} ${appendValue}`);
                }
                return;
            }
            name = line.substring(0, index);
            value = line.substring(index + 1).trim();
            // Prohibited headers
            if (/set-cookie2?/i.test(name))
                return;
            // Ignore if name is empty or some invalid character is used in name.
            if (!/^[-!#$%&'*+.^|~\d\w]+$/.test(name))
                return;
            // Check that only allowed character is used in value. Ignored if invalid.
            if (!/^[ \t\x21-\x7e\x80-\xff]*$/.test(value))
                return;
            headers.append(name, value);
        });
        return headers;
    };

    const algHash = {
        sha256: 'SHA-256',
        sha384: 'SHA-384',
        sha512: 'SHA-512',
    };
    /**
     * https://w3c.github.io/webappsec-subresource-integrity/#the-integrity-attribute
     */
    const verifyIntegrityWith = (crypto) => async (integrity, data) => {
        if (crypto?.subtle?.digest == null) {
            throw new Error('digest function is not available.');
        }
        const integrities = integrity.trim().split(/\s+/);
        for (const part of integrities) {
            const match = part.match(/^(sha(?:256|384|512))-([a-zA-Z0-9+/=]+)$/);
            if (!match) {
                throw new TypeError('unsupported alg or integrity format');
            }
            const [, alg, expected] = match;
            const hashAlgorithm = algHash[alg] || alg;
            const digest = await globalThis.crypto.subtle.digest(hashAlgorithm, await data.arrayBuffer());
            const actual = btoa(arrayBufferToString(digest));
            if (actual !== expected) {
                throw new Error('Integrity verification failed');
            }
        }
    };
    const verifyIntegrity = verifyIntegrityWith(globalThis.crypto);

    const HTTPMethods = [
        'GET',
        'POST',
        'PUT',
        'DELETE',
        'PATCH',
        'HEAD',
        'TRACE',
        'OPTIONS',
        'CONNECT',
    ];
    const isHTTPMethod = (s) => HTTPMethods.includes(s);
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const isBlob = (e) => e != null && e instanceof Blob;
    const checkHTTPMethod = (s) => {
        if (!isHTTPMethod(s)) {
            throw new Error('Not HTTP Method');
        }
        return s;
    };
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const checkBlob = (e) => {
        if (!isBlob(e)) {
            throw new Error('Not a Blob');
        }
        return e;
    };
    const buildGMHeaders = (headers) => {
        return Object.fromEntries(headers.entries());
    };
    const buildData = async (request) => {
        const arrayBuffer = await request.arrayBuffer();
        const result = arrayBufferToString(arrayBuffer);
        if (result.length === 0)
            return undefined;
        return result;
    };
    const buildWithCredentials = (request, location) => {
        const url = new URL(request.url);
        switch (request.credentials) {
            case 'include':
                return true;
            case 'omit':
                return false;
            case 'same-origin':
            default:
                return location.origin === url.origin;
        }
    };
    const buildDetails = async (request) => {
        if (request.redirect !== 'follow') {
            throw new TypeError('Only follow is supported');
        }
        const withCredentials = buildWithCredentials(request, globalThis.location);
        const headers = buildGMHeaders(request.headers);
        return {
            method: checkHTTPMethod(request.method),
            url: request.url,
            headers,
            binary: true,
            timeout: undefined,
            responseType: 'blob',
            data: await buildData(request),
            // `credentials`: whether to include cookies
            // for Tampermonkey
            anonymous: !withCredentials,
            // for GreaseMonkey
            // not documented but supported
            // https://github.com/greasemonkey/greasemonkey/issues/2826
            // https://github.com/greasemonkey/greasemonkey/pull/2835
            withCredentials,
        };
    };
    /**
     * Convert GM.Response to Response
     */
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const buildResponse = (res) => {
        const body = checkBlob(res.response);
        return new Response(body, {
            status: res.status,
            statusText: res.statusText,
            headers: parseXHRHeaders(res.responseHeaders),
        });
    };
    const gmFetch = async (resource, init = {}) => {
        if (GM.xmlHttpRequest == null) {
            throw new TypeError('GM.xmlHttpRequest is not listed in @grant.');
        }
        const request = new Request(resource, init);
        const details = await buildDetails(request);
        if (request.signal.aborted) {
            throw new DOMException('The request was aborted.', 'AbortError');
        }
        return new Promise((resolve, reject) => {
            const control = GM.xmlHttpRequest({
                ...details,
                onload(res) {
                    const checkAndBuildResponse = async () => {
                        if (request.integrity) {
                            const body = checkBlob(res.response);
                            await verifyIntegrity(request.integrity, body);
                        }
                        return buildResponse(res);
                    };
                    checkAndBuildResponse()
                        .then((response) => resolve(response))
                        // eslint-disable-next-line @typescript-eslint/prefer-promise-reject-errors
                        .catch((err) => reject(err));
                },
                onerror() {
                    reject(new TypeError('NetworkError'));
                },
                ontimeout() {
                    reject(new TypeError('TimeoutError'));
                },
                onabort() {
                    reject(new DOMException('The request was aborted.', 'AbortError'));
                },
            });
            // GM.xmlHttpRequest returns undefined in GreaseMonkey
            // but returns GMXHRControl in Tampermonkey and Violentmonkey.
            if (request.signal !== undefined) {
                if (control != null) {
                    request.signal.addEventListener('abort', () => {
                        control.abort();
                    });
                }
                else {
                    console.warn('abort is not supported in this implementation.');
                }
            }
        });
    };

    exports.gmDownload = gmDownload;
    exports.gmFetch = gmFetch;

    return exports;

})({});
